"""A workflow that re-deploys base config on a router.

When a new router is deployed, it is loaded with the current version of configuration that contain the bare necessities.
For various reasons, this template may change, and the resulting configuration follows from this. To update a router
'in the wild' where this change should be reflected, the workflow `redeploy_base_config` is used.

This workflow only takes a trouble ticket number as initial input, and deploys the base configuration, first as a dry
run. After confirmation by an operator, the configuration is committed to the machine, and this completes the workflow.
"""

from orchestrator.forms import FormPage
from orchestrator.forms.validators import Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, UUIDstr
from orchestrator.workflow import StepList, begin, done, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form

from gso.products.product_types.router import Router
from gso.services.lso_client import lso_interaction
from gso.utils.types.tt_number import TTNumber
from gso.utils.workflow_steps import deploy_base_config_dry, deploy_base_config_real


def _initial_input_form(subscription_id: UUIDstr) -> FormGenerator:
    router = Router.from_subscription(subscription_id)

    class RedeployBaseConfigForm(FormPage):
        info_label: Label = f"Redeploy base config on {router.router.router_fqdn}?"
        tt_number: TTNumber

    user_input = yield RedeployBaseConfigForm

    return user_input.model_dump() | {"subscription": router}


@workflow(
    "Redeploy base config",
    initial_input_form=wrap_modify_initial_input_form(_initial_input_form),
    target=Target.MODIFY,
)
def redeploy_base_config() -> StepList:
    """Redeploy base config on an existing router.

    * Perform a dry run of deployment
    * Redeploy base config
    """
    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> lso_interaction(deploy_base_config_dry)
        >> lso_interaction(deploy_base_config_real)
        >> resync
        >> done
    )
